import os
import re

import responses  # type: ignore


def tenant_id():
    return os.environ.get(
        'AZURE_TENANT_ID', '9b833a13-cf81-4917-a5a7-7860358faeda'  # random
    )


def mock_auth():
    endpoint = 'https://login.microsoftonline.com/' + tenant_id()
    for method in ['GET', 'POST']:
        responses.add(
            getattr(responses, method),
            re.compile(r'.*login.microsoftonline.com.*'),
            json={
                'authorization_endpoint': endpoint,
                'token_endpoint': endpoint,
                'metadata': [],
                'access_token': 'foobar',
                'refresh_token': 'foobar',
                'expires_in': 100,
                'client_info': (
                    # b64 {"uid": "foo", "utid": "bar"}
                    'eyJ1aWQiOiAiZm9vIiwgInV0aWQiOiAiYmFyIn0='
                )
            },
            status=200
        )



from abc import ABCMeta  # type: ignore
from abc import abstractmethod
from datetime import datetime
from datetime import timezone
import json
import jsonschema  # type: ignore
import os
import re
import typing as t
from urllib.parse import urlparse
from yaml import safe_load  # type: ignore

import pluggy  # type: ignore

import abnosql.exceptions as ex
from abnosql.kms import kms
from abnosql import plugin

hookimpl = pluggy.HookimplMarker('abnosql.table')
hookspec = pluggy.HookspecMarker('abnosql.table')


class TableSpecs(plugin.PluginSpec):

    @hookspec(firstresult=True)
    def set_config(self, table: str) -> t.Dict:  # type: ignore[empty-body] # noqa E501
        """Hook to set config

        Args:

            table: table name

        Returns:

            dictionary containing config

        """
        pass

    @hookspec(firstresult=True)
    def get_item_pre(self, table: str, key: t.Dict):  # type: ignore[empty-body] # noqa E501
        """Hook invoked before get_item()

        Args:

            table: table name
            key: dictionary of key passed to get_item()

        """
        pass

    @hookspec(firstresult=True)
    def get_item_post(self, table: str, item: t.Dict) -> t.Dict:  # type: ignore[empty-body] # noqa E501
        """Hook invoked after get_item()

        Args:

            table: table name
            item: dictionary item retrieved from get_item call

        Returns:

            dictionary containing updated item

        """
        pass

    @hookspec
    def put_item_pre(self, table: str, item: t.Dict) -> t.Dict:  # type: ignore[empty-body] # noqa E501
        """Hook invoked before put_item()

        Args:

            table: table name
            item: dictionary containing partition and range/sort key

        """
        pass

    @hookspec
    def put_item_post(self, table: str, item: t.Dict) -> None:  # type: ignore[empty-body] # noqa E501
        """Hook invoked after put_item()

        Args:

            table: table name
            item: dictionary containing partition and range/sort key

        """
        pass

    @hookspec
    def put_items_post(self, table: str, items: t.List[t.Dict]) -> None:  # type: ignore[empty-body] # noqa E501
        """Hook invoked after put_items()

        Args:

            table: table name
            item: list of dictionary items written to table

        """
        pass

    @hookspec
    def delete_item_post(self, table: str, key: t.Dict) -> None:  # type: ignore[empty-body] # noqa E501
        """Hook invoked after delete_item()

        Args:

            table: table name
            key: dictionary of item written to table

        """
        pass


class TableBase(metaclass=ABCMeta):

    @abstractmethod
    def __init__(
        self, pm: plugin.PM, name: str, config: t.Optional[dict] = None
    ) -> None:
        """Instantiate table object

        Args:

            pm: pluggy plugin manager
            name: table name
            config: optional config dict dict
        """
        pass

    @abstractmethod
    def get_item(self, **kwargs) -> t.Optional[t.Dict]:
        """Get table/collection item

        Args:

            partition key and range/sort key (if used)

        Returns:

            item dictionary or None if not found

        """
        pass

    @abstractmethod
    def put_item(
        self,
        item: t.Dict,
        update: t.Optional[bool] = False
    ) -> t.Dict:
        """Puts table/collection item

        Args:

            item: dictionary
            update: perform update/patch - item must already exist
            audit_user: optional user / system ID string to add audit attrs

        Returns:

            item: dictionary of created/updated item

        """
        pass

    @abstractmethod
    def put_items(
        self,
        items: t.Iterable[t.Dict],
        update: t.Optional[bool] = False,
        audit_user: t.Optional[str] = None
    ):
        """Puts multiple table/collection items

        Args:

            items: list of item dictionaries
            update: perform update/patch - items must already exist
            audit_user: user / system ID string to add audit attrs

        """
        pass

    @abstractmethod
    def delete_item(self, **kwargs):
        """Deletes table/collection item

        Args:
            partition key and range/sort key (if used)

        """
        pass

    @abstractmethod
    def query(
        self,
        key: t.Optional[t.Dict[str, t.Any]] = None,
        filters: t.Optional[t.Dict[str, t.Any]] = None,
        limit: t.Optional[int] = None,
        next: t.Optional[str] = None,
        index: t.Optional[str] = None
    ) -> t.Dict[str, t.Any]:
        """Perform key based query with optional exact match filters

        Args:

            key: dictionary containing partition key and range/sort key
            filters: optional dictionary of key=value to query and filter on
            limit: query limit
            next: pagination token
            index: name of index to use (dynamodb only)

        Returns:
            dictionary containing 'items' and 'next' pagination token

        """
        pass

    @abstractmethod
    def query_sql(
        self,
        statement: str,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        limit: t.Optional[int] = None,
        next: t.Optional[str] = None
    ) -> t.Dict[str, t.Any]:
        """Perform key based query with optional exact match filters

        Args:

            statement: SQL statement to query table
            parameters: optional dictionary containing @key = value placeholders
            limit: query limit
            next: pagination token

        Returns:
            dictionary containing 'items' and 'next' pagination token

        """
        pass


def get_sql_params(
    statement: str,
    parameters: t.Dict[str, t.Any],
    param_val: t.Callable,
    replace: t.Optional[str] = None
) -> t.Tuple[str, t.List]:
    """Get and validate statement and parameters

    Args:

        statement: SQL statement to query table
        parameters: optional dictionary containing @key = value placeholders
        param_val: callable to get
        replace: optional placeholder string (eg ?) to replace vars with

    Returns:
        statement, params

    """
    # convert @variable to dynamodb ? placeholders
    validate_statement(statement)
    vars = list(re.findall(r'\@[a-zA-Z0-9_.-]+', statement))
    params = []
    _missing = {}
    for var in vars:
        if var not in parameters:
            _missing[var] = True
        else:
            val = parameters[var]
            params.append(param_val(var, val))
    for var in parameters.keys():
        if var not in vars:
            _missing[var] = True
    missing = sorted(_missing.keys())
    if len(missing):
        raise ex.ValidationException(
            'missing parameters: ' + ', '.join(missing)
        )
    if isinstance(replace, str):
        for var in parameters.keys():
            statement = statement.replace(var, replace)
    return (statement, params)


def quote_str(string):
    # Quotes string
    return "'" + string.translate(
        string.maketrans({
            "'": "\\'"
        })
    ) + "'"


def get_item_pre(tb, kwargs):
    key = validate_key_attrs(tb.key_attrs, dict(**kwargs), False)
    tb.pm.hook.get_item_pre(table=tb.name, key=key)
    _check_exists = kwargs.pop('abnosql_check_exists', None)
    return key, _check_exists


def get_item_post(tb, kwargs, item, audit_key):
    _check_exists = kwargs.pop('abnosql_check_exists', None)
    _item = tb.pm.hook.get_item_post(table=tb.name, item=item)
    if _item:
        item = _item
    item = kms_decrypt_item(tb.config, item)
    if _check_exists is not False:
        check_exists(tb, 'get', item)
    if item is not None:
        audit_callback(tb, 'get', audit_key)
    return item


def put_item_pre(tb, item, update, audit_user):
    operation = 'update' if update else 'create'
    key = validate_key_attrs(tb.key_attrs, item)
    validate_item(tb.config, operation, item)
    item = check_exists(tb, operation, item)

    audit_user = audit_user or tb.config.get('audit_user')
    if audit_user:
        item = add_audit(item, update or False, audit_user)

    # add change metadata if enabled
    if hasattr(tb, 'change_meta') and tb.change_meta is True:
        item = add_change_meta(
            item, tb.name, 'MODIFY' if update is True else 'INSERT'
        )

    _item = tb.pm.hook.put_item_pre(table=tb.name, item=item)
    if _item:
        item = _item[0]
    item = kms_encrypt_item(tb.config, item)
    return item, key


def put_item_post(tb, item, update, audit_user, abnosql_audit_callback=True):
    key = validate_key_attrs(tb.key_attrs, item)
    tb.pm.hook.put_item_post(table=tb.name, item=item)
    if abnosql_audit_callback is not False:
        audit_callback(
            tb, 'update' if update else 'create', key, audit_user
        )
    return item


def delete_item_pre(tb, kwargs):
    key = validate_key_attrs(tb.key_attrs, dict(**kwargs), False)
    check_exists(tb, 'delete', dict(kwargs))
    return key


def delete_item_post(tb, key):
    tb.pm.hook.delete_item_post(table=tb.name, key=key)
    audit_callback(tb, 'delete', key)


def validate_query_attrs(key: t.Dict, filters: t.Dict):
    """Validate that the query and filter attributes are named correctly

    Args:

        key: key dictionary
        filters: filter dictionary

    """
    # convert @variable to dynamodb ? placeholders
    _name_pat = re.compile(r'^[a-zA-Z09_-]+$')

    def _validate_key_names(obj):
        return [_ for _ in obj.keys() if not _name_pat.match(_)]

    invalid = sorted(set(
        _validate_key_names(key) + _validate_key_names(filters)
    ))
    if len(invalid):
        raise ex.ValidationException(
            'invalid key or filter keys: ' + ', '.join(invalid)
        )


def validate_statement(statement: str):
    """Validate statement

    Args:

        statement: SQL statement

    """
    # sqlglot can do this (and sqlparse), but lets keep it simple
    tokens = [_.strip() for _ in statement.split(' ') if _.strip() != '']
    if len(tokens) == 0 or tokens[0].upper() != 'SELECT':
        raise ex.ValidationException('statement must start with SELECT')


def get_key_attrs(config: t.Optional[t.Dict] = None) -> list:
    """Get key attributes from ABNOSQL_KEY_ATTRS env var or config

    Args:

        config: config dict

    Returns:

        key_attrs: list of key attribute names

    """
    key_attrs = [
        _ for _ in os.environ.get('ABNOSQL_KEY_ATTRS', '').split(',')
        if _.strip() != ''
    ]
    _key_attrs = config.get('key_attrs') if isinstance(config, dict) else None
    if isinstance(_key_attrs, list):
        key_attrs = _key_attrs
    if len(key_attrs) > 2:
        raise ValueError('must not be more than 2 key_attrs defined')
    if len(key_attrs) == 0:
        raise ValueError('key_attrs not defined')
    return key_attrs


def validate_key_attrs(
    key_attrs: t.Iterable[str],
    item: t.Dict,
    is_put: t.Optional[bool] = True
) -> t.Dict:
    """Validate key attributes in either ABNOSQL_KEY_ATTRS env var or config

    Args:

        key_attrs: list of key attribute names
        config: config dict

    """
    if not isinstance(key_attrs, list) or len(key_attrs) == 0:
        raise ex.ValidationException('key_attrs not defined')
    missing = [
        k for k in key_attrs if item.get(k) is None
    ]
    if len(missing):
        raise ex.ValidationException(
            f'key_attrs missing from item: {missing}'
        )
    if is_put and len(item) - len(key_attrs) <= 0:
        raise ex.ValidationException(
            'item contains no additional keys beyond key_attrs'
        )
    key = {k: item[k] for k in key_attrs}
    return key


def validate_item(
    config: t.Dict, operation: str, item: t.Dict
):
    schema = config.get(f'{operation}_schema', config.get('schema'))
    if schema is None:
        return
    camel_case = os.environ.get('ABNOSQL_CAMELCASE', 'TRUE') == 'TRUE'
    meta_attr = 'changeMetadata' if camel_case else 'change_metadata'
    name_attr = 'eventName' if camel_case else 'event_name'
    existing_event = item.get(meta_attr, {}).get(name_attr)
    if existing_event == 'REMOVE':
        return
    title = config.get(
        f'{operation}_schema_errmsg',
        config.get('schema_errmsg', 'invalid item')
    )
    schema = safe_load(schema) if isinstance(schema, str) else schema
    validator = jsonschema.Draft7Validator(schema)
    errors = []
    for err in sorted(validator.iter_errors(item), key=str):
        errors.append(err.message)
    if len(errors) > 0:
        raise ex.ValidationException(title, {'errors': errors})


def kms_encrypt_item(config: t.Dict, item: t.Dict) -> t.Dict:
    """Encrypt item values as defined in config

    Each attribute value is encrypted with data key generated each time for both providers:

    - aws_kms uses aws-encryption-sdk
    - azure_kms uses Azure Keyvault RSA CMK to envelope encrypt data key

    Both providers use AESGCM generated data key with AAD/encryption context

    Example config:

        {
            'kms': {
                'key_ids': ['https://foo.vault.azure.net/keys/bar/45e36a1024a04062bd489db0d9004d09'],
                'key_attrs': ['hk', 'rk'],
                'attrs': ['obj', 'str'],
                'key_bytes': b'somekeybytearray'
            }
        }

    Where:
        - key_ids: list of AWS KMS Key ARNs or  or Azure KeyVault identifier (URL to RSA CMK).  This is picked up via `ABNOSQL_KMS_KEYS` env var as comma separated list
        - key_attrs: key attributes in the item for which to the AAD/encryption context is set
        - attrs: attributes to encrypt
        - key_bytes: use your own AESGCM key if specified, otherwise generate one

    Args:

        config: config dictionary
        item: item dict

    Returns:
        item

    """  # noqa: E501
    kcfg = config.get('kms', {})
    if item is None or not kcfg:
        return item
    context = {k: item.get(k) for k in kcfg['key_attrs']}
    # encrypt defined attrs
    for attr in kcfg['attrs']:
        val = item.get(attr)
        if val is None:
            continue
        if not isinstance(val, str):
            val = json.dumps(val)
        item[attr] = kcfg['pm'].encrypt(
            val, context, key=kcfg.get('key_bytes')
        )
    return item


def kms_decrypt_item(config: t.Dict, item: t.Dict) -> t.Dict:
    """Decrypt item as defined in config

    See kms_encrypt_item() for example config:

    Args:

        config: config dictionary
        item: item dict

    Returns:
        item
    """
    kcfg = config.get('kms', {})
    if item is None or not kcfg:
        return item
    context = {k: item.get(k) for k in kcfg['key_attrs']}
    # decrypt defined attrs
    for attr in kcfg['attrs']:
        val = item.get(attr)
        if val is None:
            continue
        val = kcfg['pm'].decrypt(val, context)
        try:
            val = json.loads(val)
        except Exception:
            val = val
        item[attr] = val
    return item


def kms_process_query_items(
    config: t.Dict,
    items: t.List[t.Dict]
) -> t.List[t.Dict]:
    """Remove encrypted attribute/values from items

    Args:

        config: config dictionary
        items: list of item dicts

    Returns:
        items

    """
    kcfg = config.get('kms')
    if not isinstance(kcfg, dict):
        return items
    _items = []
    for item in items:
        for attr in kcfg['attrs']:
            item.pop(attr, None)
        _items.append(item)
    return _items


def add_audit(item: t.Dict, update: bool, user: str) -> t.Dict:
    """Add createdBy + createdDate and/or modifiedBy + modifiedDate to item

    Args:

        item: item dict
        update: bool, true if operation is update otherwise false
        user: user/system ID string

    Returns:
        item

    """
    camel_case = os.environ.get('ABNOSQL_CAMELCASE', 'TRUE') == 'TRUE'
    by_attr = 'By' if camel_case else '_by'
    date_attr = 'Date' if camel_case else '_date'
    dt_iso = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%SZ')
    if update is not True and f'created{by_attr}' not in item:
        item.update({
            f'created{by_attr}': user,
            f'created{date_attr}': dt_iso
        })
    item.update({
        f'modified{by_attr}': user,
        f'modified{date_attr}': dt_iso
    })
    return item


def add_change_meta(item: t.Dict, event_source: str, event_name: str) -> t.Dict:
    """Add changeMetadata object to item containing eventName and eventSource

    Args:

        item: item dict
        event_source: str
        event_name: str - INSERT, MODIFY or REMOVE

    Returns:
        item

    """
    event_name = event_name.upper()
    if event_name not in ['INSERT', 'MODIFY', 'REMOVE']:
        return item
    camel_case = os.environ.get('ABNOSQL_CAMELCASE', 'TRUE') == 'TRUE'
    meta_attr = 'changeMetadata' if camel_case else 'change_metadata'
    source_attr = 'eventSource' if camel_case else 'event_source'
    name_attr = 'eventName' if camel_case else 'event_name'
    # cosmos delete_item() adds REMOVE so don't add if already present in item
    existing_event = item.get(meta_attr, {}).get(name_attr)
    if existing_event == 'REMOVE':
        return item
    item.update({
        meta_attr: {
            source_attr: event_source,
            name_attr: event_name
        }
    })
    return item


def audit_callback(tb, operation, key, audit_user=None):
    _callback = tb.config.get('audit_callback')
    audit_user = audit_user or tb.config.get('audit_user')
    if not callable(_callback) or not isinstance(audit_user, str):
        return
    dt_iso = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%SZ')
    key_str = ';'.join([f'{k}={v}' for k, v in sorted(key.items())])
    _callback(tb.name, dt_iso, operation, key_str, audit_user)


def check_exists_enabled(config):
    return config.get(
        'check_exists',
        os.environ.get('ABNOSQL_CHECK_EXISTS', 'FALSE') == 'TRUE'
    ) is True


def check_exists(obj: TableBase, operation: str, item: dict):
    if (
        len(obj.key_attrs) == 0   # type: ignore
        or not hasattr(obj, 'check_exists')
        or obj.check_exists is False
    ):
        return item
    key = {
        k: item.get(k) for k in obj.key_attrs  # type: ignore
    } if item is not None else None
    if operation == 'get' and item is None:
        raise ex.NotFoundException('item not found')
    elif operation == 'create' and key is not None:
        # can be overridden if defined in item
        if item.pop('abnosql_check_exists', None) is False:
            return item
        key['abnosql_check_exists'] = False
        if obj.get_item(**key) is not None:
            raise ex.ExistsException('item already exists')
    elif operation == 'update' and key is not None:
        if obj.get_item(**key) is None:
            raise ex.NotFoundException('item not found')
    elif operation == 'delete' and key is not None:
        if obj.get_item(**key) is None:
            raise ex.NotFoundException('item not found')
    return item


def parse_connstr():
    connstr = os.environ.get('ABNOSQL_DB')
    if connstr:
        return urlparse(connstr)
    return None


def table(
    name: str,
    config: t.Optional[dict] = None,
    database: t.Optional[str] = None
) -> TableBase:
    """Create table object

    Args:

        name: table name
        config: optional config
        database: optional database

    Returns:
        TableBase object

    """
    if database is None:
        p = parse_connstr()
        database = p.scheme or p.path if p else None

    # infer database from available env vars
    # aws: https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html  # noqa
    # azure: https://learn.microsoft.com/en-us/azure/azure-functions/functions-app-settings#azure_functions_environment  # noqa
    if database is None:
        defaults = {
            'AWS_DEFAULT_REGION': 'dynamodb',
            'FUNCTIONS_WORKER_RUNTIME': 'cosmos',
            'K_SERVICE': 'firestore'
        }
        for envvar, _database in defaults.items():
            if os.environ.get(envvar) is not None:
                database = _database
                break

    if database is None:
        raise ex.PluginException('table plugin database not defined')

    pm = plugin.get_pm('table')
    module = pm.get_plugin(database)
    if module is None:
        raise ex.PluginException(f'table.{database} plugin not found')
    if hasattr(module, 'MISSING_DEPS'):
        raise ex.PluginException(
            f'table.{database} plugin missing dependencies'
        )
    if not isinstance(config, dict):
        config = {}
    _module = module.Table(pm, name, config)

    # load crypto module
    kcfg = config.get('kms')
    if isinstance(kcfg, dict):
        defaults = {
            'dynamodb': 'aws',
            'cosmos': 'azure',
            'firestore': 'gcp'
        }
        provider = kcfg.get('provider')
        if database is not None and provider is None:
            provider = defaults.get(database)
        _kms_module = kms(kcfg, provider)
        config['kms']['pm'] = _kms_module

        if 'key_attrs' not in kcfg:
            config['kms']['key_attrs'] = get_key_attrs(config)

        if 'session' in config and 'session' not in kcfg:
            # aws_encryption_sdk uses botocore session
            config['kms']['session'] = config['session']._session

        # check required attrs
        missing = [
            _ for _ in ['attrs', 'key_attrs']
            if not isinstance(kcfg.get(_), list) or len(kcfg[_]) == 0
        ]
        if len(missing):
            raise ex.ConfigException(
                'kms config missing %s' % ', '.join(missing)
            )

    return _module



from base64 import b64decode
from base64 import b64encode
import functools
import json
import os
import typing as t

import abnosql.exceptions as ex
from abnosql.kms import get_keys
from abnosql.kms import KmsBase
from abnosql.kms import pack_bytes
from abnosql.kms import unpack_bytes
from abnosql.plugin import PM


try:
    import azure.core.exceptions as azex  # type: ignore
    from azure.identity import DefaultAzureCredential  # type: ignore
    from azure.keyvault.keys.crypto import CryptographyClient  # type: ignore
    from azure.keyvault.keys.crypto import KeyWrapAlgorithm  # type: ignore
    from cryptography.hazmat.primitives.ciphers.aead import AESGCM
except ImportError:
    MISSING_DEPS = True


def kms_ex_handler(raise_not_found: t.Optional[bool] = True):

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except azex.ResourceNotFoundError as e:
                if raise_not_found:
                    raise ex.NotFoundException(detail=e.message) from None
                return None
            except Exception as e:
                raise ex.PluginException(detail=e)
        return wrapper
    return decorator


class Kms(KmsBase):

    def __init__(
        self, pm: PM, config: t.Optional[dict] = None
    ) -> None:
        self.pm = pm
        self.config = config or {}
        self.provider = 'azure'
        key_ids = self.config.get('key_ids', get_keys())
        if not isinstance(key_ids, list) or len(key_ids) == 0:
            raise ex.ConfigException('kms key_ids required')
        self.key_id = key_ids[0]
        self.crypto_client = CryptographyClient(
            self.key_id, self.config.get(
                'credential', DefaultAzureCredential()
            )
        )
        self.pack_bytes_maxlen = self.config.get(
            'pack_bytes_maxlen', 10000
        )

    @kms_ex_handler()
    def encrypt(
        self, plaintext: str, context: t.Dict, key: t.Optional[bytes] = None
    ) -> str:
        # azure doesnt have GenerateDataKey equivilent
        # as AWS does, and its encrypt/decrypt APIs
        # are only for use against CMKs not data keys
        # so generate own AESGCM DEK to encrypt/decrypt
        # the plaintext and then use azure to wrap/unwrap this with CMK.
        # This follows similar pattern to aws-encryption-sdk and google tink
        # and the wrapped/encrypted AES key lives with the data
        # see https://developers.google.com/tink/client-side-encryption
        # https://docs.aws.amazon.com/encryption-sdk/latest/developer-guide/concepts.html  # noqa
        context = dict(sorted(context.items()))
        aad = json.dumps(context).encode()

        # 1) generate random Data Encryption Key (DEK)
        # 256-bit AES-GCM key with 96-bit nonce
        nonce = os.urandom(96)
        dek = key or AESGCM.generate_key(bit_length=256)
        dek_aesgcm = AESGCM(dek)

        # 2) The DEK is encrypted by a Key Encryption Key (KEK)
        # that is stored in a cloud KMS (Azure Key Vault CMK)
        enc_dek = self.crypto_client.wrap_key(
            KeyWrapAlgorithm.rsa_oaep_256, dek
        ).encrypted_key
        del dek  # delete unencrypted DEK from memory asap

        # 3) Data is encrypted using the DEK by the client.
        ct = dek_aesgcm.encrypt(nonce, plaintext.encode(), aad)
        del dek_aesgcm

        # 4) Concatenates the KEK-encrypted encryption DEK with the encrypted
        # data (byte packing is what aws-encryption-sdk and google tink do)
        serialized = b64encode(
            pack_bytes([ct, nonce, enc_dek], self.pack_bytes_maxlen)
        ).decode()
        return serialized

    @kms_ex_handler()
    def decrypt(self, serialized: str, context: t.Dict) -> str:
        context = dict(sorted(context.items()))
        aad = json.dumps(context).encode()

        # 1) Extracts the KEK-encrypted DEK key.
        unpacked = unpack_bytes(b64decode(serialized.encode()))
        if len(unpacked) != 3:
            raise ValueError('invalid serialization')
        (ct, nonce, enc_dek) = unpacked

        # 2) Makes a request to your KMS to decrypt the KEK-encrypted DEK.
        # decrypt the key using Azure Key Vault CMK
        dek = self.crypto_client.unwrap_key(
            KeyWrapAlgorithm.rsa_oaep_256, enc_dek
        ).key

        # 3) Decrypts the ciphertext locally using the DEK.
        dek_aesgcm = AESGCM(dek)
        del dek
        plaintext = dek_aesgcm.decrypt(nonce, ct, aad).decode()
        return plaintext



from base64 import b64decode
from base64 import b64encode
import functools
import json
import os
import typing as t

import abnosql.exceptions as ex
from abnosql.kms import get_keys
from abnosql.kms import KmsBase
from abnosql.plugin import PM

try:
    from tink import aead  # type: ignore
    from tink import core  # type: ignore
    from tink.integration import gcpkms  # type: ignore
    from tink import new_keyset_handle  # type: ignore
except ImportError:
    MISSING_DEPS = True


def kms_ex_handler(raise_not_found: t.Optional[bool] = True):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except core.TinkError as e:
                raise ex.ConfigException(detail=str(e))
            except Exception as e:
                raise ex.PluginException(detail=e)
        return wrapper
    return decorator


def mock_remote_aead(*args, **kwargs):
    # used for patching during tests
    # see https://github.com/tink-crypto/tink-py/blob/main/tink/aead/_kms_envelope_aead_test.py  # noqa
    keyset_handle = new_keyset_handle(aead.aead_key_templates.AES256_GCM)
    return keyset_handle.primitive(aead.Aead)


class Kms(KmsBase):

    @kms_ex_handler()
    def __init__(
        self, pm: PM, config: t.Optional[dict] = None
    ) -> None:
        self.pm = pm
        self.config = config or {}
        self.provider = 'gcp'

        self.key_ids = self.config.get('key_ids', get_keys())
        if not isinstance(self.key_ids, list) or len(self.key_ids) == 0:
            raise ex.ConfigException('kms key_ids required')
        self.kek_uri = self.key_ids[0]
        self.credentials = self.config.get(
            'credentials', os.environ.get('GOOGLE_APPLICATION_CREDENTIALS')
        )
        # see https://developers.google.com/tink/client-side-encryption
        aead.register()
        self.client = gcpkms.GcpKmsClient(
            self.kek_uri,
            self.credentials
        )
        remote_aead = self.client.get_aead(self.kek_uri)
        self.env_aead = aead.KmsEnvelopeAead(
            aead.aead_key_templates.AES256_GCM, remote_aead
        )

    @kms_ex_handler()
    def encrypt(
        self, plaintext: str, context: t.Dict, key: t.Optional[bytes] = None
    ) -> str:
        ciphertext = self.env_aead.encrypt(
            plaintext.encode(), json.dumps(context).encode()
        )
        return b64encode(ciphertext).decode()

    @kms_ex_handler()
    def decrypt(self, serialized: str, context: t.Dict) -> str:
        plaintext = self.env_aead.decrypt(
            b64decode(serialized), json.dumps(context).encode()
        )
        return plaintext.decode()



from base64 import b64decode
from base64 import b64encode
import functools
import json
import logging
import os
from os.path import expanduser
import sys

import typing as t

import pluggy  # type: ignore

import abnosql.exceptions as ex
from abnosql.plugin import PM
from abnosql.table import check_exists_enabled
from abnosql.table import delete_item_post
from abnosql.table import delete_item_pre
from abnosql.table import get_item_post
from abnosql.table import get_item_pre
from abnosql.table import get_key_attrs
from abnosql.table import kms_process_query_items
from abnosql.table import parse_connstr
from abnosql.table import put_item_post
from abnosql.table import put_item_pre
from abnosql.table import TableBase
from abnosql.table import validate_query_attrs

import sqlglot
from sqlglot import exp

hookimpl = pluggy.HookimplMarker('abnosql.table')

try:
    from google.api_core.exceptions import ClientError  # type: ignore
    from google.auth.exceptions import GoogleAuthError  # type: ignore
    from google.cloud import firestore  # type: ignore
except ImportError:
    MISSING_DEPS = True

OPERATORS = {
    exp.EQ: '==',
    exp.NEQ: '!=',
    exp.GT: '>',
    exp.GTE: '>=',
    exp.LT: '<',
    exp.LTE: '<=',
}


def firestore_ex_handler(raise_not_found: t.Optional[bool] = True):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ClientError as e:
                if raise_not_found and e.code in [404]:
                    raise ex.NotFoundException() from None
                raise ex.ValidationException(detail=e) from None
            except GoogleAuthError as e:
                raise ex.ConfigException(detail=e) from None
            except ex.NoSQLException:
                raise
            except Exception as e:
                raise ex.PluginException(detail=e) from None
        return wrapper
    return decorator


class Table(TableBase):

    @firestore_ex_handler()
    def __init__(
        self, pm: PM, name: str, config: t.Optional[dict] = None
    ) -> None:
        self.pm = pm
        self.name = name
        self.database = 'firestore'
        self.set_config(config)
        self.client = self.config.get('client', self.get_client())
        self.key_attrs = get_key_attrs(self.config)
        self.check_exists = check_exists_enabled(self.config)
        self.table = self.client.collection(name)
        self.docid_delim = self.config.get('docid_delim', ':')
        self.batch = None

    def get_client(self):
        kwargs = {}
        pc = parse_connstr()
        if kwargs is not None:
            kwargs.update({
                'project': pc.username,
                'database': pc.hostname
            })
            if pc.password not in ['', None]:
                kwargs['credentials'] = pc.password
        for attr in ['project', 'database', 'credentials']:
            for prefix in ['ABNOSQL_FIRESTORE_', 'GOOGLE_CLOUD_']:
                val = os.environ.get(prefix + attr.upper(), kwargs.get(attr))
                if val is not None:
                    kwargs[attr] = val
                break
        cred_file = os.path.join(*(
            (
                tuple(os.environ.get('APPDATA', ''))
                if sys.platform == 'win32'
                else (expanduser('~'), '.config')
            ) + (
                'gcloud',
                'application_default_credentials.json'
            )
        ))
        gac = 'GOOGLE_APPLICATION_CREDENTIALS'
        if gac not in os.environ and os.path.isfile(cred_file):
            os.environ[gac] = cred_file
        return firestore.Client(**kwargs)

    def _docid(self, **kwargs):
        item = dict(kwargs)
        key = [
            item.get(attr)
            for attr in self.key_attrs
            if item.get(attr)
        ]
        if len(key) > 2 or len(key) == 0:
            raise ValueError('key length must be 1 or 2')
        return self.docid_delim.join(key)

    @firestore_ex_handler()
    def set_config(self, config: t.Optional[dict]):
        if config is None:
            config = {}
        _config = self.pm.hook.set_config(table=self.name)
        if _config:
            config = t.cast(t.Dict, _config)
        self.config = config

    @firestore_ex_handler()
    def get_item(self, **kwargs) -> t.Optional[t.Dict]:
        audit_key, _ = get_item_pre(self, dict(**kwargs))

        doc = self.table.document(self._docid(**kwargs)).get()
        item = doc.to_dict() if doc.exists else None

        return get_item_post(self, dict(**kwargs), item, audit_key)

    @firestore_ex_handler()
    def put_item(
        self,
        item: t.Dict,
        update: t.Optional[bool] = False,
        audit_user: t.Optional[str] = None
    ) -> t.Dict:
        item, _ = put_item_pre(self, item, update, audit_user)

        # do update
        docid = self._docid(**item)
        ref = self.table.document(docid)
        if update is True:
            if self.batch:
                self.batch.update(ref, item)
            else:
                ref.update(item)

        # do create/replace
        else:
            if self.batch:
                self.batch.set(ref, item)
            else:
                ref.set(item)

        # firestore doesnt return updated item, so make this optional if needed
        # note encrypted attrs won't be decrypted
        if self.config.get('put_get') is True:
            item = self.table.document(docid).get().to_dict()

        return put_item_post(self, item, update, audit_user)

    @firestore_ex_handler()
    def put_items(
        self,
        items: t.Iterable[t.Dict],
        update: t.Optional[bool] = False,
        audit_user: t.Optional[str] = None
    ):
        if self.config.get('batchmode') is not False:
            self.batch = self.client.batch()
        for item in items:
            self.put_item(item, update=update, audit_user=audit_user)
        self.pm.hook.put_items_post(table=self.name, items=items)
        if self.config.get('batchmode') is not False:
            if self.batch is not None:
                self.batch.commit()
            self.batch = None

    @firestore_ex_handler()
    def delete_item(self, **kwargs):
        key = delete_item_pre(self, dict(kwargs))

        docid = self._docid(**kwargs)
        self.table.document(docid).delete()
        delete_item_post(self, key)

    @firestore_ex_handler()
    def query(
        self,
        key: t.Optional[t.Dict[str, t.Any]] = None,
        filters: t.Optional[t.Dict[str, t.Any]] = None,
        limit: t.Optional[int] = None,
        next: t.Optional[str] = None,
        index: t.Optional[str] = None
    ) -> t.Dict[str, t.Any]:
        filters = filters or {}
        key = key or {}
        validate_query_attrs(key, filters)
        parameters = {
            f'@{k}': v for k, v in
            (filters | key).items()
        }
        statement = 'SELECT * FROM table'
        op = 'WHERE'
        for param in parameters.keys():
            statement += f' {op} {param} = @{param}'
            op = 'AND'
        return self.query_sql(
            statement,
            parameters,
            limit=limit,
            next=next
        )

    @firestore_ex_handler()
    def query_sql(
        self,
        statement: str,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        limit: t.Optional[int] = None,
        next: t.Optional[str] = None
    ) -> t.Dict[str, t.Any]:
        limit = limit or 100
        parameters = parameters or {}
        select = None
        try:
            select = sqlglot.parse_one(statement)
        except Exception:
            raise ex.ValidationException(detail='invalid SQL')
        if not isinstance(select, exp.Select):
            raise ex.ValidationException(detail='only SELECT is supported')

        # parse the sql using sqlglot (there must be a better way to do below)
        filters = []
        query = self.table
        where = select.find(exp.Where)
        if where:
            for cond in where.find_all(exp.Condition):
                if type(cond) not in OPERATORS.keys():
                    continue
                column = cond.this.name
                expr = cond.expression
                operator = OPERATORS.get(type(cond))  # type: ignore
                val = expr.this.name
                pval = parameters.get(f'@{val}')
                if isinstance(expr, exp.Column):
                    filters.append([column, operator, val])
                elif isinstance(expr, exp.Parameter) and pval is not None:
                    filters.append([column, operator, pval])

        logging.debug(f'query_sql() table: {self.name}, filters: {filters}')
        query = self.table

        for (col, op, val) in filters:
            query = query.where(col, op, val)

        # don't order as it messes up pagination
        if next is not None:
            query = query.start_at(
                json.loads(b64decode(next).decode())
            )

        if limit is not None:
            # needs to be + 1 so can see if any more left
            # as firestore doesnt tell us if anything left to paginate
            # so have to peak ahead with limit + 1
            query = query.limit(limit + 1)

        c = 0
        items = []
        last = None
        for doc in query.stream():
            c += 1
            item = doc.to_dict()
            if c < limit + 1:
                items.append(item)
            last = b64encode(json.dumps({
                k: item[k] for k in self.key_attrs
                if k in item
            }).encode()).decode()

        if c < limit + 1:
            last = None
        items = kms_process_query_items(self.config, items)
        return {
            'items': items,
            'next': last
        }



import logging
from traceback import format_tb


class NoSQLException(Exception):

    def __init__(self, title=None, detail=None, status=None):
        super().__init__(title)
        self.title = title
        self.detail = detail
        self.status = status

    def to_problem(self):
        obj = {}
        if isinstance(self.detail, Exception):
            obj = {
                'exception': str(self.detail),
                'stackTrace': format_tb(self.detail.__traceback__)
            }
            logging.error(self.title, obj)
        return {
            'title': self.title,
            'detail': obj.get('exception', self.detail),
            'status': self.status,
            'type': None
        }


class ExistsException(NoSQLException):
    def __init__(self, title=None, detail=None, status=409):
        super(ExistsException, self).__init__(
            title or 'already exists', detail, status
        )


class NotFoundException(NoSQLException):
    def __init__(self, title=None, detail=None, status=404):
        super(NotFoundException, self).__init__(
            title or 'not found', detail, status
        )


class ConfigException(NoSQLException):
    def __init__(self, title=None, detail=None, status=500):
        super(ConfigException, self).__init__(
            title or 'invalid config', detail, status
        )


class PluginException(NoSQLException):
    def __init__(self, title=None, detail=None, status=500):
        super(PluginException, self).__init__(
            title or 'plugin exception', detail, status
        )


class ValidationException(NoSQLException):
    def __init__(self, title=None, detail=None, status=400):
        super(ValidationException, self).__init__(
            title or 'validation exception', detail, status
        )
